FlipMatrix <- function(X) {
  n <- nrow(X)
  Y <- X
  for (i in 1:n) {
    Y[, i] <- X[, n + 1 - i]
  }
  return(Y)
}

FindCut <- function(Y, s, a) {
  #class specific coverage classification region
  Y0 <- Y[Y==0]
  Y1 <- Y[Y==1]
  n0 <- length(Y0)
  n1 <- length(Y1)
  s0 <- s[Y==0]
  s1 <- s[Y==1]
  t0 <- sort(s0, decreasing = T)[floor(n0 * a[1])]
  t1 <- sort(s1, decreasing = F)[floor(n1 * a[2])]
  ambi <- mean(s >= t1 & s <= t0)
  return(t=c(t0,t1,ambi))
}

MinCut <- function(Y, s, a, tol = 1e-3) {
  # total covarege
  pi0 <- mean(Y==0)
  pi1 <- 1 - pi0
  a.start <- 0.1 * a / pi0
  a.end <- 0.9 * a / pi0
  a.vec <- seq(start = a.start, end = a.end, by = a / pi0 / 100)
  ambi.vec <- zeros(length(a.vec))
  for (i in 1:length(a.vec)) {
    a0 <- a.vec[i]
    a1 <- (a - a0 * pi0) / pi1
    ambi.vec[i] <- FindCut(Y, s, c(a0, a1))[3]
  }
  a0.opt <- a.vec[which(ambi.vec < min(ambi.vec) + tol)]
  a0.opt1 <- mean(a.opt)
  a1.opt1 <- (a - a0.opt1 * pi0) / pi1
  return(list(cut = FindCut(Y, s, c(a0.opt1, a1.opt1)), a.opt = c(a0.opt1, a1.opt1)))
}

SearchD <- function(s, rej.prop) {
  d.set <- c(0,sort(abs(s-0.5)))
  n.d <- length(d.set)
  n.rej <- 0
  ii <- 1
  while (n.rej < rej.prop) {
    ii <- ii + 1
    d.tmp <- d.set[ii]
    n.rej <- length(which(abs(s-0.5)<=d.tmp))
  }
  return(d.tmp)
}

PopCov <- function(t0, t1, mu, sd.vec, p) {
  return(list(miss.0 = 1 - pnorm(t0, mu[1], sd.vec[1]),
              miss.1 = pnorm(t1, mu[2], sd.vec[2]),
              ambi = (pnorm(t0, mean = mu[1], sd = sd.vec[1]) - 
                      pnorm(t1, mean = mu[1], sd = sd.vec[1])) * p[1] +
                     (pnorm(t0, mean = mu[2], sd = sd.vec[2]) - 
                      pnorm(t1, mean = mu[2], sd = sd.vec[2])) * p[2]
              )
        )
}

PerformClass <- function(Y.new, s.new, t) {
  k <- nrow(s.new)
  miss0 <- rep(0, k)
  miss1 <- rep(0, k)
  ambi <- rep(0, k)
  for (i in 1:k) {
    miss0[i] <- sum(Y.new == 0 & s.new[i, ] > t[1]) / sum(Y.new == 0)
    miss1[i] <- sum(Y.new == 1 & s.new[i, ] < t[2]) / sum(Y.new == 1)
    ambi[i] <- sum(s.new[i, ] >= t[2] & s.new[i, ] <= t[1]) / length(Y.new)
  }
  return(data.frame(miss0 = miss0, miss1 = miss1, ambi = ambi))
}

CV.AmbiClass <- function(Y, X, nfold = 3, a = c(0.01, 0.01)) {
  n <- length(Y)
  rand.ind <- sample.int(n, n)
  rank.sample.size <- floor(n / nfold)
  rank.ind <- rand.ind[1:rank.sample.size]
  fit.ind <- rand.ind[(rank.sample.size + 1):n]
  Y.fit <- Y[fit.ind]
  X.fit <- X[fit.ind, ]
  Y.rank <- Y[rank.ind]
  X.rank <- X[rank.ind, ]
  ind.rank.0 <- which(Y.rank == 0)
  ind.rank.1 <- which(Y.rank == 1)
  n0.rank <- length(ind.rank.0)
  n1.rank <- length(ind.rank.1)
  my.glm <- glmnet(x = X.fit, y = Y.fit, family = 'binomial', standardize = F)
  n.lambda <- length(my.glm$lambda)
  cut.values <- matrix(0, ncol = 2, nrow = n.lambda)
  ambi.vec <- rep(0, n.lambda)
  for (i in 1:n.lambda) {
    s.new <- predict(my.glm, newx = X.rank, s = my.glm$lambda[i], type = 'response')
    s.new.0 <- s.new[ind.rank.0]
    s.new.1 <- s.new[ind.rank.1]
    t0 <- sort(s.new.0, decreasing = T)[max(floor(n0.rank * a[1]), 1)]
    t1 <- sort(s.new.1, decreasing = F)[max(floor(n1.rank * a[2]), 1)]
    cut.values[i, 1] <- t0
    cut.values[i, 2] <- t1
    ambi.vec[i] <- sum(s.new <= t0 & s.new >= t1) / rank.sample.size
  }
  return(list(cut.values = cut.values, ambi = ambi.vec, glm.fit <- my.glm))
}

Find.KNN.Ind <- function(dist.vec, k) {
  n <- length(dist.vec)
  res.ind <- 1:k
  res.val <- dist.vec[res.ind]
  max.val <- max(res.val)
  for (i in (k+1):n) {
    if (dist.vec[i] < max.val) {
      res.ind <- c(setdiff(res.ind, res.ind[which.max(res.val)]), i)
      res.val <- dist.vec[res.ind]
      max.val <- max(res.val)
    }
  }
  return(res.ind)
}

KNN.Label <- function(X, Y, k = 1, X.new) {
  knn.ind <- get.knnx(X, X.new, k)$nn.index
  return(matrix(Y[knn.ind], ncol = ncol(knn.ind)))
}

KNN.Conf.Mult <- function(X, Y, k, X.test = NULL, alpha = NULL, rank.prop = 0, fill.white = 'n') {
  n <- length(Y)
  n.test <- nrow(X.test)
  class.labels <- sort(unique(Y))
  n.class <- length(class.labels)
  if (is.null(alpha)) {alpha <- rep(0.05, n.class)}
  # split the training sample to fitting and ranking subsamples
  if (rank.prop == 0) { # If proportion is not specified, then conformalization is not used.
    X.fit <- X
    X.rank <- X
    Y.fit <- Y
    Y.rank <- Y
  } else {
    rand.ind <- sample.int(n, n)
    rank.sample.size <- floor(n * rank.prop)
    rank.ind <- rand.ind[1:rank.sample.size]
    fit.ind <- rand.ind[(rank.sample.size + 1):n]
    Y.fit <- Y[fit.ind]
    X.fit <- X[fit.ind,]
    Y.rank <- Y[rank.ind]
    X.rank <- X[rank.ind,]
  }
  
  if (is.null(X.test)) {
    X.test <- X.rank
  }
  
  # get k nearest neighbors in fitting subsample for each point in ranking subsample
  knn.label.mat <- KNN.Label(X.fit, Y.fit, k, X.rank) # each row is the label of k nn for a ranking subsample point
  
  # gives for each point in ranking subsample, the proporition of labels in k neaerest neighbors
  knn.prop.rank <- lapply(1:n.class,
    function(i){
      knn.label.i <- knn.label.mat[which(Y.rank == class.labels[i]), ]
      apply(knn.label.i ==  class.labels[i], 1, mean)}) # this is a list of length 10. The k'th list entry contains the proportions of label 'k-1' among K-nn, for all ranking subsample points whose true label is 'k-1'.
  
  # now do Knn on the test sample
  knn.label.mat.test <- KNN.Label(X.fit, Y.fit, k, X.test) # labels of K nearest neighbors in fitting subsample, for points in the test sample
  knn.prop.mat.test <- 
    t(apply(knn.label.mat.test, 1,
            function(x){table(factor(x, levels = class.labels)) / k})) # proportion of labels in Knn for each point in the test sample ( \hat p( k | x_{new} ) for each k )
  knn.rank.test <-
    t(apply(knn.prop.mat.test, 1,
      function(x){
        mapply(function(y,z){(sum(z<=y)+1)/(length(z)+1)}, x, knn.prop.rank)
      })) # these are the ranking among the ranking subsample scores
  knn.conf.class.test <-
    t(apply(knn.rank.test, 1,
      function(x){
        as.numeric(x >= alpha)
      })) # conformal classifier indicator
      
  # now fill the null region with some method 
  knn.conf.class.test.fill <- knn.conf.class.test
  tholds <- NULL # tholds can be computed with Get.Threshold unless greedy filling used
  if (fill.white != 'n' & min(apply(knn.conf.class.test.fill, 1, sum)) < 1) {
	if (fill.white == 'greedy'){
		# t_0: thresholds from original solution
		# we will decrease them greedily until they are <= 1
		t_0 <- sapply( 1:length(knn.prop.rank), 
			function(i){
				score.vec <- sort(knn.prop.rank[[i]])
				n.score <- length(score.vec)
				score.vec[floor((n.score + 1) * alpha[i])]
			})
		
		max_t_decrease <- 100 # maximum number of decrements for each threshold
		potential.tholds <- lapply(t_0,seq,to=0,length.out=max_t_decrease+2)
		potential.tholds <- lapply(potential.tholds,
									FUN=function(x) x[2:(max_t_decrease+1)]) # thold cannot be 0
		i <- rep(0,n.class) # indicates number of decrements made to each threshold
		classes <- 1:n.class
		t_can_decrease <- t_0 > 0
		H_star <- t(apply(knn.prop.mat.test, 1, # current LABEL classifier in test sample
			function(x){
				as.numeric(x >= t_0)
			}))
		ambiguity <- mean(rowSums(H_star))
		# start greedy decrements
		t_iter <- t_0
		iter=1
		set.seed(1234)
		while(sum(t_iter)>1){
			iter=iter+1
			# potential new values of the tholds
			t_candidates <- mapply(function(x,y) x[y], potential.tholds, i+1)
			# if a thold cannot decrease then it is fixed at current value
			t_candidates[!t_can_decrease] <- t_iter[!t_can_decrease]
			
			H_candidates <-
				t(apply(knn.prop.mat.test, 1,
				function(x){
					as.numeric(x >= t_candidates)
				})) # this would be the new LABEL using t_candidates
			
			# these are the n.class possible values of ambiguity from changing one thold at a time
			amb_cand <- ambiguity[iter-1] - colMeans(H_star) + colMeans(H_candidates)
			amb_cand <- amb_cand[t_can_decrease]
			
			# pick the thold that increases ambiguity the least
			k_greedy <- which(amb_cand==min(amb_cand))
			k_greedy <- sample(x=k_greedy,size=1)
			k_greedy <- classes[t_can_decrease][k_greedy]
			
			t_iter[k_greedy] <- t_candidates[k_greedy]
			
			i[k_greedy] <- i[k_greedy] + 1
			
			t_can_decrease[k_greedy] <- i[k_greedy] < max_t_decrease
				
			H_star[,k_greedy] <- H_candidates[,k_greedy]
			ambiguity[iter] <- mean(rowSums(H_star))
		}
		tholds <- t_iter
		knn.conf.class.test.fill <- H_star
	} else {
		ind.no <- which(apply(knn.conf.class.test, 1, sum) < 1)
		if (fill.white == 'next') {
		fill.class <- apply(knn.rank.test[ind.no,], 1, which.is.max)
		}
		if (fill.white == 'bayes') {
		fill.class <- apply(knn.prop.mat.test[ind.no,], 1, which.is.max)
		}
		for (ii in 1:length(ind.no)) {
		knn.conf.class.test.fill[ind.no[ii], fill.class[ii]] <- 1
		}
	}
  }
  return(list(ranking.score = knn.prop.rank, ranking = knn.rank.test, class = knn.conf.class.test,
              class.fill = knn.conf.class.test.fill, tholds=tholds))
}


Get.Threshold <- function(ranking.score, alpha) {
  sapply(1:length(ranking.score), 
         function(i){
           score.vec <- sort(ranking.score[[i]])
           n.score <- length(score.vec)
           score.vec[floor((n.score + 1) * alpha[i])]
         })
}

Eval.Conf.Mult <- function(H.res, Y.test, class.labels) {
  n.test <- length(Y.test)
  pred.size <- apply(H.res, 1, sum)
  ind.null <- which(pred.size == 0)
  coverage.ind <- sapply(1:n.test, function(i){H.res[i, which(class.labels==Y.test[i])]})
  coverage.vec <- tapply(coverage.ind, Y.test, mean)
  no.pred.vec <- table(factor(Y.test[ind.null], levels = class.labels)) /
                 table(factor(Y.test, levels = class.labels))
  ambiguity <- mean(pred.size)
  co.occur <- t(H.res) %*% H.res
  return(list(coverage=coverage.vec, null.proportion =no.pred.vec, ambi = ambiguity, co.occur=co.occur))
}

CV.KNN <- function(X,Y,k.vec,n.fold=3) {
  n <- nrow(X)
  folds <- createFolds(1:n, n.fold)
  ind.star <- which.max(sapply(k.vec,function(k){sum(sapply(1:n.fold,function(v){
    test.fold <- folds[[v]]
    train.fold <-setdiff(1:n,test.fold)
    sum(knn(X[train.fold,], X[test.fold,],Y[train.fold],k)==Y[test.fold])
  }))}))
  return(k.vec[ind.star])
}

Plot.Zip <- function(X, Y, file.name = 'zip_image.eps') {
  digit.list <- sort(unique(Y))
  n.digit <- length(digit.list)
  digit.count <- rep(0, n.digit)
  digit.ind <- list()
  for (j in 1:n.digit) {
    digit.tmp <- digit.list[j]
    digit.ind[[j]] <- which(Y == digit.tmp)
    digit.count[j] <- length(digit.ind[[j]])  
  }
  max.length <- max(digit.count)
  #pdf('ambi_image.pdf', width = 6, height = 4)
  setEPS()
  postscript(file.name,  width = 6, height = 4)
  par(mfrow = c(n.digit, max.length), mar = rep(0.1,4))
  for (j in 1:n.digit) {
    for (l in 1:max.length) {
      if (l <= digit.count[j]) {
        image(FlipMatrix(matrix(X[digit.ind[[j]][l],], ncol = 16)),
              xaxt = 'n', yaxt = 'n', col = gray(0:255/255))
      } else {
        plot.new()
      }
    }
  }
  dev.off()
}